import * as mc from "@minecraft/server"
import { MessageFormData, uiManager } from "@minecraft/server-ui"
import { world, system } from "@minecraft/server"
import Utility from "./Utility"
import ChestFormData from "./ChestForms"
import ItemDatabase from "./ItemDatabase"
const { ForceOpen } = Utility

const PlayerStatus = {
  NOTRADE: "notrade",
  TRADE_MENU: "trade_menu",
  TRADE_SELECTING_ITEM: "trade_selecting_item",
  READY: "trade_ready",
  CONFIRM_MENU: "trade_confirm",
  ACCEPT: "trade_accept"
}

export default class TradeClass {
  /**
   * @private
   * @type {{ [key: string]: { isTrading: boolean, status: PlayerStatus otherPlayer?: mc.Player } }}
   */
  #PlayerTrades = {};

  /**
   * @private
   * @type {{ player1: mc.Player, player2: mc.Player, createdAt: number }[]}
   */
  #Trades = [];

  constructor() {
    this.Database = new ItemDatabase("Trade1.0.0")
    world.beforeEvents.playerLeave.subscribe(({ player }) => {
      const tradeData = this.getTradeData(player)
      if (!tradeData) return
      const oppositePlayer = this.#PlayerTrades[player.name].otherPlayer

      this.#PlayerTrades[player.name] = {
        isTrading: false,
        status: PlayerStatus.NOTRADE
      }
      this.#PlayerTrades[oppositePlayer.name] = {
        isTrading: false,
        status: PlayerStatus.NOTRADE
      }

      system.run(() => {
        if (oppositePlayer.isValid()) {
          oppositePlayer.sendMessage("§eTrade canceled because other player leaved.")
          uiManager.closeAllForms(oppositePlayer)
          this.Database.getItemsByData({ owner: oppositePlayer.name }).forEach(i => oppositePlayer.dimension.spawnItem(i.unStore(false), oppositePlayer.location))
        }
        if (this.#Trades.findIndex(t => t == tradeData) > -1) this.#Trades.splice(this.#Trades.findIndex(t => t == tradeData), 1)
      })
    })

    world.afterEvents.playerSpawn.subscribe(async ({ player, initialSpawn }) => {
      await this.Database.waitLoaded()
      const items = this.Database.getItemsByData({ owner: player.name })
      console.warn(`${player.name} has ${items.length} item(s) in database.`)
      if (initialSpawn && items.length > 0) {
        const LoginUI = new MessageFormData()
          .title("Trade Log")
          .body("Its looks like you got logout when doing trading. Click on any button to restore your items!")
          .button2("Okay")
          .button1("Close")

        ForceOpen(player, LoginUI).then(() => {
          this.Database.getItemsByData({ owner: player.name }).forEach(i => player.dimension.spawnItem(i.unStore(false), player.location))
          player.sendMessage("§aItems successfully restored!")
        })
      }
    })

    world.afterEvents.worldInitialize.subscribe(() => {
      let running = false
      system.runInterval(() => {
        if (running) return
        running = true
        for (const tradeData of this.#Trades) {
          const { player1, player2 } = tradeData
          if ((this.#PlayerTrades[player1.name].status == PlayerStatus.ACCEPT) && (this.#PlayerTrades[player2.name].status == PlayerStatus.ACCEPT)) {
            this.Database.getItemsByData({ owner: player1.name }).forEach(i => player2.dimension.spawnItem(i.unStore(false), player2.location))
            this.Database.getItemsByData({ owner: player2.name }).forEach(i => player1.dimension.spawnItem(i.unStore(false), player1.location))

            this.#PlayerTrades[player1.name] = {
              isTrading: false,
              status: PlayerStatus.NOTRADE
            }
            this.#PlayerTrades[player2.name] = {
              isTrading: false,
              status: PlayerStatus.NOTRADE
            }

            if (this.#Trades.findIndex(t => t == tradeData) > -1) this.#Trades.splice(this.#Trades.findIndex(t => t == tradeData), 1)
            player1.sendMessage("§aThe trade was successful.")
            player2.sendMessage("§aThe trade was successful.")
          }
        }
        running = false
      })
    })
  }

  /**
   * @param {mc.Player} player 
   */
  getTradeData(player) {
    return this.#Trades.find(d => d.player1 == player || d.player2 == player)
  }

  /**
  * @param {mc.Player} player 
  */
  isInTrade(player) {
    return this.getTradeData(player) != undefined
  }

  /**
   * @param {mc.Player} player 
   * @param {this.#Trades} data
   */
  getPlayerNum(player, data, opposite = false) {
    if (data.player1 == player) return opposite ? "player2" : "player1"
    if (data.player2 == player) return opposite ? "player1" : "player2"
  }


  /**
   * @param {mc.Player} player 
   * @returns {boolean}
   */
  isReady(player) {
    return this.#PlayerTrades[player.name].status == PlayerStatus.READY
  }

  /**
   * @param {mc.Player} player 
   * @param {boolean} ready
   */
  setReady(player, ready) {
    this.#PlayerTrades[player.name].status = ready ? PlayerStatus.READY : PlayerStatus.TRADE_MENU
  }

  /**
   * @param {mc.Player} player 
   * @param {mc.Player} targetPlayer 
   */
  sendTrade(player, targetPlayer) {
    const confirmUI = new MessageFormData()
      .title("Trade Request")
      .body(`${player.name} sent you request to trade!`)
      .button2("Accept")
      .button1("Decline")

    player.sendMessage("§eTrade request succesfully sent.")
    targetPlayer.sendMessage(`§e${player.name} sent you trade request.`)
    ForceOpen(targetPlayer, confirmUI).then((res) => {
      if (!res.canceled) {
        if (res.selection == 1) {
          if (!player.isValid() || !targetPlayer.isValid()) return
          if (this.isInTrade(player) || this.isInTrade(targetPlayer)) return
          this.#PlayerTrades[player.name] = {
            isTrading: true,
            status: PlayerStatus.TRADE_MENU,
            otherPlayer: targetPlayer
          }
          this.#PlayerTrades[targetPlayer.name] = {
            isTrading: true,
            status: PlayerStatus.TRADE_MENU,
            otherPlayer: player
          }
          this.#Trades.push({
            player1: player,
            player2: targetPlayer,
            createdAt: Date.now()
          })
          this.showTradeMenu(player)
          this.showTradeMenu(targetPlayer)
          player.sendMessage(`§a${targetPlayer.name} accepted your request.`)
        } else {
          player.sendMessage(`§c${targetPlayer.name} declined your request.`)
        }
      }
    })
  }

  /**
   * @param {mc.Player} player 
   */
  showTradeMenu(player) {
    const tradeData = this.getTradeData(player)
    const playerState = this.#PlayerTrades[player.name]
    if (!tradeData || playerState.status == PlayerStatus.NOTRADE) return
    if (playerState.status == PlayerStatus.CONFIRM_MENU) return
    const Player1 = tradeData.player1
    const Player2 = tradeData.player2
    const oppositePlayer = playerState.otherPlayer
    const TradeUI = new ChestFormData(`trade_${player == Player1 ? "left" : "right"}`)
      .title(oppositePlayer.name.slice(0, 10))

    const Player1Slot = [
      0, 1, 2,
      9, 10, 11,
      18, 19, 20
    ]

    const Player2Slot = [
      6, 7, 8,
      15, 16, 17,
      24, 25, 26,
    ]

    const Player1Items = this.Database.getItemsByData({ owner: Player1.name })
    for (let i = 0; i < 9; i++) {
      const itemData = Player1Items[i]
      if (itemData) {
        const item = itemData.item
        let itemDesc = []
        let enchantments = item.getComponent("minecraft:enchantable")?.getEnchantments() ?? []
        let enchantedItem = false
        if (enchantments.length > 0) {
          enchantedItem = true
          enchantments.forEach(e => itemDesc.push(Utility.enchantToText(e)))
        }
        itemDesc.push("")
        item.getLore().forEach(l => itemDesc.push("§o§5" + l))
        itemDesc.push(`§cClick to remove from trade`)
        TradeUI.button(Player1Slot[i], `${Utility.getItemname(item)} §r(${item.typeId})`, itemDesc, item.typeId, item.amount, enchantedItem)
      }
    }

    const Player2Items = this.Database.getItemsByData({ owner: Player2.name })
    for (let i = 0; i < 9; i++) {
      const itemData = Player2Items[i]
      if (itemData) {
        const item = itemData.item
        let itemDesc = []
        let enchantments = item.getComponent("minecraft:enchantable")?.getEnchantments() ?? []
        let enchantedItem = false
        if (enchantments.length > 0) {
          enchantedItem = true
          enchantments.forEach(e => itemDesc.push(Utility.enchantToText(e)))
        }
        itemDesc.push("")
        item.getLore().forEach(l => itemDesc.push("§o§5" + l))
        itemDesc.push(`§cClick to remove from trade`)
        TradeUI.button(Player2Slot[i], `${Utility.getItemname(item)} §r(${item.typeId})`, itemDesc, item.typeId, item.amount, enchantedItem)
      }
    }

    TradeUI.button(40, "Add Item", null, "textures/items/nether_star")

    TradeUI.button(45, "Refresh", null, "minecraft:wind_charge")

    let is1Ready = this.isReady(Player1)
    TradeUI.button(48, `${Player1.name} (${is1Ready ? "Ready" : "Not Ready"})`, null, `textures/blocks/glass_${is1Ready ? "lime" : "red"}`)

    TradeUI.button(49, "Cancel", null, "textures/blocks/barrier")

    let is2Ready = this.isReady(Player2)
    TradeUI.button(50, `${Player2.name} (${is2Ready ? "Ready" : "Not Ready"})`, null, `textures/blocks/glass_${is2Ready ? "lime" : "red"}`)

    ForceOpen(player, TradeUI).then((res) => {
      if (!res.canceled) {
        if (Player1Slot.includes(res.selection) && player == Player1) { // Remove Item Player 1
          const slot = Player1Slot.findIndex(i => i == res.selection)
          const item = Player1Items[slot]
          if (item.isValid()) player.dimension.spawnItem(item.unStore(false), player.location)
          this.setReady(player, false)
          if (this.#PlayerTrades[Player2.name].status != PlayerStatus.TRADE_SELECTING_ITEM) {
            this.setReady(Player2, false)
            uiManager.closeAllForms(Player2)
          }

        } else if (Player2Slot.includes(res.selection) && player == Player2) { // Remove Item Player 2
          const slot = Player2Slot.findIndex(i => i == res.selection)
          const item = Player2Items[slot]
          if (item.isValid()) player.dimension.spawnItem(item.unStore(false), player.location)
          this.setReady(player, false)
          if (this.#PlayerTrades[Player1.name].status != PlayerStatus.TRADE_SELECTING_ITEM) {
            this.setReady(Player1, false)
            uiManager.closeAllForms(Player1)
          }

        } else if (res.selection == 40) { // Adding Item
          if (this.Database.getAmountByData({ owner: player.name }) < 9) {
            if (this.#PlayerTrades[oppositePlayer.name].status != PlayerStatus.TRADE_SELECTING_ITEM && this.isReady(player)) uiManager.closeAllForms(oppositePlayer)
            this.#PlayerTrades[player.name].status = PlayerStatus.TRADE_SELECTING_ITEM
            return this.showSelectItem(player)
          }

        } else if (res.selection == 48 && player == Player1) { // Ready Player 1
          this.setReady(player, !this.isReady(player))
          if (this.isReady(player) && this.isReady(oppositePlayer)) {
            this.#PlayerTrades[player.name].status = PlayerStatus.CONFIRM_MENU
            this.#PlayerTrades[oppositePlayer.name].status = PlayerStatus.CONFIRM_MENU
            uiManager.closeAllForms(oppositePlayer)
            this.showConfirmUI(oppositePlayer)
            return this.showConfirmUI(player)
          }
          if (this.#PlayerTrades[oppositePlayer.name].status != PlayerStatus.TRADE_SELECTING_ITEM) uiManager.closeAllForms(oppositePlayer)

        } else if (res.selection == 49) { // Cancel
          return this.closeTrade(player)

        } else if (res.selection == 50 && player == Player2) { // Ready Player 2
          this.setReady(player, !this.isReady(player))
          if (this.isReady(player) && this.isReady(oppositePlayer)) {
            this.#PlayerTrades[player.name].status = PlayerStatus.CONFIRM_MENU
            this.#PlayerTrades[oppositePlayer.name].status = PlayerStatus.CONFIRM_MENU
            uiManager.closeAllForms(oppositePlayer)
            this.showConfirmUI(oppositePlayer)
            return this.showConfirmUI(player)
          }
          if (this.#PlayerTrades[oppositePlayer.name].status != PlayerStatus.TRADE_SELECTING_ITEM) uiManager.closeAllForms(oppositePlayer)

        }
      }

      return this.showTradeMenu(player)
    })
  }

  /**
   * @param {mc.Player} player 
   */
  showSelectItem(player) {
    const tradeData = this.getTradeData(player)
    const playerState = this.#PlayerTrades[player.name]
    if (!tradeData || playerState.status == PlayerStatus.NOTRADE) return
    const oppositePlayer = playerState.otherPlayer
    const SelectItemUI = new ChestFormData("inv")
      .title("Select item")
      .button(49, "Back", null, "textures/blocks/barrier")

    const playerInventory = player.getComponent("inventory").container
    for (let slot = 0; slot < playerInventory.size; slot++) {
      let item = playerInventory.getItem(slot)
      if (!item) continue
      let itemDesc = []
      let enchantments = item.getComponent("minecraft:enchantable")?.getEnchantments() ?? []
      let enchantedItem = false
      if (enchantments.length > 0) {
        enchantedItem = true
        enchantments.forEach(e => itemDesc.push(Utility.enchantToText(e)))
      }
      itemDesc.push("")
      item.getLore().forEach(l => itemDesc.push("§o§5" + l))
      itemDesc.push(`§eClick to select the item`)
      SelectItemUI.button(slot, Utility.getItemname(item), itemDesc, item.typeId, item.amount, enchantedItem)
    }

    for (let i = 36; i < 45; i++) { SelectItemUI.button(i, "", null, "textures/blocks/glass_black") }

    ForceOpen(player, SelectItemUI).then((res) => {
      if (!res.canceled) {
        if (res.selection < 36) {
          const item = playerInventory.getItem(res.selection)
          if (item) {
            playerInventory.setItem(res.selection)
            this.Database.add(item, { owner: player.name })
            this.setReady(player, false)
            if (this.isReady(oppositePlayer)) this.setReady(oppositePlayer, false)
            if (this.#PlayerTrades[oppositePlayer.name].status != PlayerStatus.TRADE_SELECTING_ITEM) uiManager.closeAllForms(oppositePlayer)
          }
        } else if (res.selection == 49) {
          return this.showTradeMenu(player)
        } else {
          return this.showSelectItem(player)
        }
      }

      return this.showTradeMenu(player)
    })
  }

  showConfirmUI(player) {
    const tradeData = this.getTradeData(player)
    const playerState = this.#PlayerTrades[player.name]
    if (!tradeData || playerState.status == PlayerStatus.NOTRADE) return
    const oppositePlayer = playerState.otherPlayer
    const items = this.Database.getItemsByData({ owner: oppositePlayer.name })
    const confirmUI = new MessageFormData()
      .title("Confirmation UI")
      .body(`Are you sure to accept trade with ${oppositePlayer.name}
You will get:
${items.length > 0 ? items.map(i => `${i.item.typeId} x${i.item.amount}`).join("\n") : "Nothing"}`)
      .button2("Accept")
      .button1("Cancel")

    ForceOpen(player, confirmUI).then((res) => {
      if (!res.canceled) {
        if (res.selection == 1) {
          this.#PlayerTrades[player.name].status = PlayerStatus.ACCEPT
        } else {
          return this.closeTrade(player)
        }
      } else {
        return this.showConfirmUI(player)
      }
    })
  }

  /**
   * @param {mc.Player} player 
   */
  closeTrade(player) {
    const tradeData = this.getTradeData(player)
    if (!tradeData) return
    const Player1 = tradeData.player1
    const Player2 = tradeData.player2
    this.#PlayerTrades[player.name].otherPlayer.sendMessage(`§c${player.name} canceled the trade.`)
    this.#PlayerTrades[Player1.name] = {
      isTrading: false,
      status: PlayerStatus.NOTRADE
    }
    this.#PlayerTrades[Player2.name] = {
      isTrading: false,
      status: PlayerStatus.NOTRADE
    }
    uiManager.closeAllForms(Player1)
    this.Database.getItemsByData({ owner: Player1.name }).forEach(i => Player1.dimension.spawnItem(i.unStore(false), Player1.location))
    uiManager.closeAllForms(Player2)
    this.Database.getItemsByData({ owner: Player2.name }).forEach(i => Player1.dimension.spawnItem(i.unStore(false), Player2.location))
    if (this.#Trades.findIndex(t => t == tradeData) > -1) this.#Trades.splice(this.#Trades.findIndex(t => t == tradeData), 1)
  }

}
