import * as mc from "@minecraft/server"
import { world, system } from "@minecraft/server"
import { ActionFormData } from "@minecraft/server-ui"
import TradeClass from "./Modules/Trade"
import Restful from "./Modules/Restful"
import CommandBuilder from "./Modules/CommandBuilder"
import Config from "./Configuration"
import Utility from "./Modules/Utility"
const { ForceOpen } = Utility

const Version = "1.0.0"
const Trade = new TradeClass()

const warn = console.warn
console.warn = (...args) => {
  return warn(`[Trade System v${Version}]`, ...args)
}
const error = console.error
console.error = (...args) => {
  return error(`[Trade System v${Version} | ERROR]`, ...args)
}

world.beforeEvents.worldInitialize.subscribe(({ itemComponentRegistry }) => {
  console.warn("Loading system...")
  itemComponentRegistry.registerCustomComponent("pao:trade_item_onuse", {
    onUse: ({ itemStack, source: player }) => {
      console.warn(`${player.name} Opening trade menu...`)
      player.runCommand("scriptevent pao_trade:initial")
    }
  })
})

/**
 * @param {mc.Player} player 
 */
const openTradeMenu = (player) => {
  console.warn(`${player.name} Trade menu opened.`)
  const players = world.getAllPlayers().filter(p => p != player)
  const tradeUI = new ActionFormData()
    .title("Trade Menu")
    .body("Select player")

  for (const p of players) {
    tradeUI.button(p.name)
  }
  tradeUI.button("§c<< Close >>")

  ForceOpen(player, tradeUI).then((res) => {
    if (!res.canceled) {
      if (res.selection != players.length) {
        /**
             * @type {mc.Player}
             */
        const targetPlayer = players[res.selection]
        if (!targetPlayer.isValid()) return player.sendMessage("§cPlayer not found.")

        player.runCommand(`scriptevent tradesystem:runCommand trade "${targetPlayer.name}"`)
      }
    }
  })
}

const Command = new CommandBuilder("tradesystem")
world.beforeEvents.chatSend.subscribe(async (data) => {
  let player = data.sender
  let message = data.message
  let prefix = Config.CommandPrefix
  if (message.startsWith(prefix)) {
    let args = data.message.slice(prefix.length).trim().split(' ')
    let command = args.shift().toLowerCase()
    let cmd = Command.getAllRegistation().find(c => c.name == command || (c.aliases && c.aliases.includes(command)))
    if (!cmd) return
    data.cancel = true
    system.run(() => {
      player.runCommandAsync(`scriptevent ${cmd.id}:runCommand ${command} ${args.join(" ")}`)
    })
  }
})

Command.afterCommandUse.subscribe(({ error, player, command }) => {
  if (command.id != "tradesystem") return
  if (error) {
    player.sendMessage(`§c${error}${error?.stack ?? ""}`)
  }
})

Command.register({
  name: "trade",
  description: "Send trade request to a player",
  category: "Trade",
  usage: "trade <player_name?>"
}, (data, player, args) => {
  if (player.hasTag("ban-trade")) return
  if (Date.now() < (player.getDynamicProperty("tradeCooldown") ?? Date.now())) return player.sendMessage("§eYou just use trade, try again later!")
  if (!args[0]) {
    player.sendMessage("§eClose Chat to show UI")
    return player.runCommand("scriptevent pao_trade:initial")
  }
  const targetPlayer = world.getAllPlayers().find(p => p.name == args[0])
  if (!targetPlayer || !targetPlayer.isValid()) return player.sendMessage("§cPlayer not found.")

  if (Trade.isInTrade(targetPlayer)) return player.sendMessage("§cPlayer is in trade.")
  if (Trade.isInTrade(player)) return player.sendMessage("§cYou are in trade.")

  player.setDynamicProperty("tradeCooldown", Date.now() + (Config.TradeCooldown * 1000))

  return Trade.sendTrade(player, targetPlayer)
})

world.afterEvents.worldInitialize.subscribe(() => {
  system.afterEvents.scriptEventReceive.subscribe(({ id, sourceEntity: player, message }) => {
    if (!player || !player.isValid() || player.typeId != "minecraft:player") return
    switch (id.toLowerCase()) {
      case "pao_trade:initial": return openTradeMenu(player)
      case "pao_trade:log": return console.warn(message)
    }
  }, { "namespaces": ["pao_trade"] })

  Restful.request("mce-installed").then(async (data) => {
    if (data?.installed) {
      console.warn("Minecraft Essentials installed.")
      Restful.request("registerCreditToMCE", { addonName: "Trade System", creator: "Paoeni", version: Version })
      Restful.request("registerMenuToMCE", { buttonDisplay: "Trade", buttonImage: "textures/items/emerald", direct: { scriptEvent: "pao_trade:initial" } })
    }
  })

  Restful.listen("tradesystem-installed", () => {
    return { installed: true, version: Version }
  })

  console.warn("Loaded.")
})